// Viewer.java
// Justin Ludiwg 1999.03.01

package edu.wooster.lu99jp26.link;

import com.docuverse.dom.DOM;
import com.jclark.xsl.sax.Driver;
import edu.wooster.lu99jp26.util.Function;
import java.awt.Component;
import java.awt.Frame;
import java.awt.MenuItem;
import java.awt.Point;
import java.awt.PopupMenu;
import java.awt.Rectangle;
import java.awt.Window;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.IOException;
import java.lang.Exception;
import java.lang.Integer;
import java.lang.NullPointerException;
import java.lang.String;
import java.lang.System;
import java.lang.Throwable;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLConnection;
import java.util.Properties;
import java.util.Vector;
import javax.swing.JEditorPane;
import javax.swing.event.HyperlinkEvent;
import javax.swing.event.HyperlinkListener;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.w3c.dom.ProcessingInstruction;

public class Viewer extends JEditorPane
implements ActionListener, HyperlinkListener, MouseListener
{
	// one-way linked-list of viewers
	// each has an unique id number
	protected static Viewer first = null;
	protected static int count = 0;
	protected static int newCount = 0;
	
	protected Viewer next;
	protected int number;
	
	// vector of open xlinks
	protected static Vector xlinks = new Vector (100, 100);
	// vector of unprocessed xpointers
	protected static Vector xpointers = new Vector (100, 100);
	
	protected static DOM dom = null;
	
	protected Document doc;
	protected URL url;
	protected String stylesheet = "";
	protected boolean getNullPage = false;
	
	// keep track of mouseDowns so we'll know when to popup menu
	// (requires a double-click & hold)
	protected Point mouse;
	protected boolean showPopup = false;
	protected PopupMenu popup;
	
	
	// -- static methods --

    /**
     * Returns first Viewer in list with an URL of url, or null if none
     * @param url URL
     * @return Viewer
     */
	public static Viewer viewerWithURL (URL url)
	{
		Viewer current = first;
		
		while (current != null)
		{
			if (current.url.equals (url))
				return current;
			
			current = current.next;
		}
		
		return null;
	}

    /**
     * Returns first Viewer in list with a Document of doc, or null if none
     * @param doc Document
     * @return Viewer
     */
	public static Viewer viewerWithDoc (Document doc)
	{
		Viewer current = first;
		
		while (current != null)
		{
			if (current.doc.equals (doc))
				return current;
			
			current = current.next;
		}
		
		return null;
	}
	
	// -- instance methods --
	
	/**
	 * creates a blank new Viewer
	 * @param dom DOM the DOM to use, or null
	 */
	public Viewer ()
	{
		super ();
		
		setEditable (false);
		addHyperlinkListener (this);
		addMouseListener (this);
		
		// create a new DOM if this is the first Viewer opened
		if (dom == null)
		{
			Properties p = System.getProperties ();
			p.put ("org.xml.sax.parser", "edu.wooster.lu99jp26.link.SAXParser");
			System.setProperties (p);
			
			dom = new DOM ();
			dom.setWriter (new DOMWriterImpl ());
		}
		
		popup = new PopupMenu ();
		add (popup);
		
		addViewer ();
	}
	
	/**
	 * prepares this Viewer for extinction
	 */
	public void close ()
	{
		
	}
	
	// -- linked-list methods --

    /**
     * Adds this to linked list; sets next and number
     */
	protected synchronized void addViewer ()
	{
		// make new viewer first in list
		Viewer second = first;
		first = this;
		next = second;
		
		// increment current number of Viewers in list
		count++;
		
		// give this Viewer unique number
		number = ++newCount;
	}

    /**
     * Removes this from linked list; sets next and number
     */
	protected synchronized void removeViewer ()
	{
		// remove this Viewer from list
		if (first == this)
			first = next;
		else
		{
			Viewer current = first.next;
			Viewer previous = first;
			
			// walk thru list until we find previous Viewer
			while (current != this)
			{
				previous = current;
				current = current.next;
			}
			
			previous.next = next;
		}
		
		// deccrement current number of Viewers in list
		count--;
	}

    /**
     * Returns true if there are Viewers after this one in the list
     */
	protected synchronized boolean moreViewers ()
	{
		return next != null;
	}

    /**
     * Returns first Viewer in list, or null if none
     */
	protected synchronized Viewer firstViewer ()
	{
		return first;
	}

    /**
     * Returns next Viewer in list, or null if none
     */
	protected synchronized Viewer nextViewer ()
	{
		return next;
	}
	
	// -- event handlers --

    /**
     * Invoked when an action occurs
     * use this to trap the menu events from the popup menu
     */
	public void actionPerformed (ActionEvent e)
	{
		String s = e.getActionCommand ();
		
		int linkNum = Integer.valueOf (s.substring (0, s.indexOf ('-')))
			.intValue ();
    	int locatorNum = Integer.valueOf (s.substring (s.indexOf ('-') + 1))
    		.intValue ();
    	Element link = (Element) xlinks.elementAt (linkNum);
    	
    	try
    	{
	    	if (locatorNum == 0) try
	    	{
	    		setPage (new URL (viewerWithDoc (link.getOwnerDocument ())
	    				.url.toString () + "#" + linkNum + "-0").toString ());
	    	}
			catch (NullPointerException npe) {}
			
			else if (locatorNum == 1 && 
					 link.getAttribute ("xlink:form").equals ("simple"))
			{
				String href = link.getAttribute ("xlink:href");
				
		    	if (XPointer.getXPointer (href).equals (""))
					setPage (new URL (url, href).toString ());
				
				else
					setPage (new URL (url, XPointer.stripXPointer (href) 
							+ "#" + linkNum + "-1").toString ());
			}
			else // extended
			{
				NodeList locators = link.getChildNodes ();
				int numlocators = locators.getLength (), locatorCount = 0;
				
				for (int k = 0; k < numlocators; k++)
				{
					Node n = locators.item (k);
					if (n instanceof Element)
					{
						Element f = (Element) n;
						if (f.getAttribute ("xlink:form").equals ("locator"))
						{
							if (locatorNum == ++locatorCount)
							{
								String href = f.getAttribute ("xlink:href");
								
						    	if (XPointer.getXPointer (href).equals (""))
									setPage (new URL (url, href).toString ());
								
								else
									setPage (new URL (url, 
									XPointer.stripXPointer (href) + "#" 
									+ linkNum + "-" + locatorNum).toString ());
							}
						}
					}
				}
			}
		}
		catch (IOException ioe) {}
	}
	
	/**
     * Invoked when a hypertext link is updated.
     */
    public void hyperlinkUpdate (HyperlinkEvent e)
    {
    	URL url = e.getURL ();
    	String type = e.getEventType ().toString ();
    	
    	/* never seem to get these two
    	if (type.equals ("ENTERED"))
    	{}
    	else if (type.equals ("EXITED"))
    	{}
    	else*/ if (type.equals ("ACTIVATED"))
    	{
    		popup.removeAll ();
			String s = url.toString ();
    		
    		if (s.startsWith ("http://xlink/~"))
    		{
    			int linkNum = Integer.valueOf (s.substring 
    				(s.lastIndexOf ('~') + 1, s.lastIndexOf ('-'))).intValue ();
    			int locatorNum = Integer.valueOf (s.substring 
    				(s.lastIndexOf ('-') + 1)).intValue ();
    			Element link = (Element) xlinks.elementAt (linkNum);
    			
    			if (link.getAttribute ("xlink:form").equals ("simple"))
    			{
    				if (locatorNum == 0)
    				{
    					String title = link.getAttribute ("xlink:title");
    					if (title.equals (""))
    						title = link.getAttribute ("xlink:href");
	    				
	    				addPopupItem (title, linkNum + "-1");
    				}
    				else if (locatorNum == 1) try
	    			{
	    				String title = link.getAttribute ("xlink:content-title");
	    				if (title.equals (""))
	    					title = viewerWithDoc (link.getOwnerDocument ())
	    						.url.toString ();
	    				
	    				addPopupItem (title, linkNum + "-0");
	    			}
	    			catch (NullPointerException npe) {}
	    		}
	    		else // extended
	    		{
	    			// if the extended element is part of the link,
	    			// add an item for it
	    			if (locatorNum > 0 && 
	    				link.getAttribute ("xlink:inline").equals ("true")) try
	    			{
	    				String title = link.getAttribute ("xlink:content-title");
	    				if (title.equals (""))
	    					title = viewerWithDoc (link.getOwnerDocument ())
	    						.url.toString ();
	    				
	    				addPopupItem (title, linkNum + "-0");
	    			}
	    			catch (NullPointerException npe) {}
	    			
	    			// add the each locator (except for locatorNum)
	    			NodeList locators = link.getChildNodes ();
	    			int numlocators = locators.getLength ();
	    			int locatorCount = 0;
	    			
	    			for (int k = 0; k < numlocators; k++)
	    			{
	    				Node n = locators.item (k);
	    				if (n instanceof Element)
	    				{
	    					Element f = (Element) n;
	    					if (f.getAttribute ("xlink:form").equals ("locator") 
	    						&& ++locatorCount != locatorNum)
	    					{
	    						String title = f.getAttribute ("xlink:title");
	    						if (title.equals (""))
	    							title = f.getAttribute ("xlink:href");
	    						
	    						addPopupItem (title, linkNum + "-" + locatorCount);
	    					}
	    				}
	    			}
	    		}
	    		
	    		// show the popup on next mouseDown (if double-click & hold)
	    		showPopup = true;
    		}
    		else
    			setPage (s);
    	}
    }
    
    // -- mouse event handlers --
    
    /**
     * Invoked when the mouse has been clicked on a component.
     */
    public void mouseClicked (MouseEvent e) {}

    /**
     * Invoked when a mouse button has been pressed on a component.
     * if double-click & hold on a link, show the popup
     */
    public void mousePressed (MouseEvent e)
    {
    	int fudge = 2;
    	Point newMouse = e.getPoint ();
    	
    	if (showPopup
    		&& newMouse.x + fudge > mouse.x && newMouse.x - fudge < mouse.x 
    		&& newMouse.y + fudge > mouse.y && newMouse.y - fudge < mouse.y)
    		popup.show (this, newMouse.x + 5, newMouse.y - 5);
    	
    	mouse = newMouse;
    	showPopup = false;
    }

    /**
     * Invoked when a mouse button has been released on a component.
     */
    public void mouseReleased (MouseEvent e) {}

    /**
     * Invoked when the mouse enters a component.
     */
    public void mouseEntered (MouseEvent e) {}

    /**
     * Invoked when the mouse exits a component.
     */
    public void mouseExited (MouseEvent e) {}
    
    // -- overloaded JEditorPane methods --
    
    /**
	 * passes URL in String form
	 * @param url String the URL
	 */
	public void setPage (String url)
	{
		setPage (url, true);
	}
    
    
    /**
	 * passes URL in URL form
	 * @param url URL the URL
	 */
	public void setPage (URL url)
	{
		setPage (url.toString (), true);
	}
    
    /**
	 * passes URL in String form
	 * @param url String the URL
	 * @param autoDisplay boolean if true, must call displayPage () to display
	 */
	public void setPage (String url, boolean autoDisplay)
	{
		try
		{
			String noXPointer = XPointer.stripXPointer (url);
			// the file w/ xpointer and w/o, respectively
			URL fileRef = new URL (this.url, url);
			URL absoluteFile = new URL (this.url, noXPointer);
			
	    	if (this.url == null)
			{
				// set the title of this window to the url of this document
				Component window = Function.getRootContainer (this);
				if (window instanceof Frame)
				{
					((Frame) window).setTitle (noXPointer);
					((Frame) window).show ();
				}
				
				Progress.setString ("opening " + url);
				this.url = absoluteFile;
				
				try
				{
					if (noXPointer.endsWith (".xml"))
					{
						addLinks ();
						
						if (autoDisplay)
							super.setPage (fileRef);
					}
					else
						super.setPage (fileRef);
				}
				catch (Exception e)
				{
					Progress.setString ("failed to open " + url);
					e.printStackTrace ();
				}
				
				Progress.setString ("");
			}
			// just scroll to named anchor
			else if (absoluteFile.equals (this.url)) try
			{
				super.setPage (fileRef);
			}
			catch (IOException ioe) {}
			
			// try matching the url to other open documents
			else try
			{
				Viewer remote = viewerWithURL (absoluteFile);
				remote.setPage (url);
				
				Component window = Function.getRootContainer (remote);
				if (window instanceof Window)
					((Window) window).show ();
			}
			// if the page hasn't already been opened, open a new window
			catch (NullPointerException npe)
			{
				ViewerFactory.newDefaultViewer (url);
			}
		}
		catch (MalformedURLException murle) {}
	}

	/**
	 * Fetch a stream for the given url, which is about to 
	 * be loaded by the <code>setPage</code> method.  By
	 * default, this simply opens the url and returns the
	 * stream.  This can be reimplemented to do useful things
	 * like fetch the stream from a cache, monitor the progress
	 * of the stream, etc.
	 * <p>
	 * This method is expected to have the the side effect of
	 * establising the content type, and therefore setting the
	 * appropriate EditorKit to use for loading the stream.
	 */
	protected InputStream getStream (URL page) throws IOException
	{
		URLConnection conn = url.openConnection ();
		String type = conn.getContentType ();
		
		// guess content from extension
		if (type.startsWith ("content/unknown"))
		{
			if (url.toString ().endsWith (".xml"))
				type = "text/xml";
			
			else if (url.toString ().endsWith (".html") ||
					 url.toString ().endsWith (".htm"))
				type = "text/html";
			
			else if (url.toString ().endsWith (".text") ||
					 url.toString ().endsWith (".txt"))
				type = "text/plain";
			
			else if (url.toString ().endsWith (".jpeg") ||
					 url.toString ().endsWith (".jpg"))
				type = "image/jpeg";
			
			else if (url.toString ().endsWith (".gif"))
				type = "image/gif";
		}
		
		if (type.startsWith ("text/html") || type.startsWith ("text/plain"))
		{
			Progress.setString ("rendering html");
			setContentType (type);
			return conn.getInputStream ();
		}
		
		// work our xml magic
		else if (type.startsWith ("text/xml"))
		{
			processLinks ();
			processStylesheet ();
			
			Progress.setString ("rendering html " + url.toString ());
			setContentType ("text/html");
			
			return new FileInputStream ("data/temp/result" + number + ".html");
		}
		
		// use html template to produce html file w/ image tag
		else if (type.startsWith ("image/jpeg") || type.startsWith ("image/gif"))
		{
			processImage ();
			
			Progress.setString ("rendering image");
			setContentType ("text/html");
			return new FileInputStream ("data/temp/image.html");
		}
		
		throw new IOException ("unknown type: " + type);
	}

    /**
     * Gets the current url being displayed.  If a URL was 
     * not specified in the creation of the document, this
     * will return null, and relative URL's will not be 
     * resolved.
     *
     * @return the URL
     */
    public URL getPage()
    {
    	if (getNullPage)
    		return null;
    	
    	return super.getPage ();
    }
	
	// -- XML processing functions --
	
	/**
	 * Processes image template file to contain the url of this image
	 * replaces the string "^1" in image.html with the url string
	 */
	protected void processImage () throws IOException
	{
		Progress.setString ("processing image");
		
		InputStream template = new BufferedInputStream 
			(new FileInputStream ("data/image.html"));
		OutputStream result = new BufferedOutputStream 
			(new FileOutputStream ("data/temp/image.html"));
		
		int c = template.read ();
		
		// do until end of file
		while (c != -1)
		{
			// copy template into result file until special '^' encountered
			if (c == '^')
			{
				// read item key (i.e. ^1 or ^2 or ^3 etc.)
				c = template.read ();
				
				// write page url
				if (c == '1')
				{
					result.write (url.toString ().getBytes ());
				}
			}
			else
				result.write (c);
			
			c = template.read ();
		}
		
		template.close ();
		result.flush ();
		result.close ();
	}
	
	/**
	 * Adds link tags in this file to xlinks and xpointers Vectors
	 * alsa deals with group links
	 * and reprocesses open files to which new xpointers apply
	 */
	protected void addLinks () throws IOException
	{
		doc = dom.openDocument (url);
		if (doc == null)
			throw new IOException (dom.getException ().toString ());
		
		Progress.setString ("adding xlinks " + url.toString ());
		
		Element docElem = doc.getDocumentElement ();
		docElem.setAttribute ("xmlns:html", "http://www.w3.org/TR/REC-html40");
		
		Vector viewers = new Vector (5, 5);
		addLinks (docElem, viewers);
		
		for (int i = 0; i < viewers.size (); i++) try
		{
			String href = (String) viewers.elementAt (i);
			Viewer remote = viewerWithURL (new URL (url, href));
			
			if (remote != null)
				remote.displayPage ();
		}
		catch (Exception e) {}
	}
	
	/**
	 * adds link tags from this new doc to xlinks vector
	 * also processes inline links
	 */
	protected void addLinks (Element e, Vector viewers)
	{
		// if form != "", is an xlink element
		String form = e.getAttribute ("xlink:form");
		
		if (form.equals ("simple") || form.equals ("extended"))
		{
			// remote anchor href="http://xlink/~[link#]-1" name="[link#]-1"
			if (form.equals ("simple"))
				addSimpleLink (e, viewers);
			
			// remote anchors href="http://xlink/~[link#]-[locator#]"
			// name="[link#]-[locator#]"
			else if (form.equals ("extended"))
				addExtendedLink (e, viewers);
			
			// local anchor href="http://xlink/~[link#]-0" name="[link#]-0"
			if (e.getAttribute ("xlink:inline").equals ("true"))
				addAnchor (e, xlinks.size (), 0);
			
			xlinks.addElement (e);
		}
		else if (form.equals ("group"))
			addGroupLink (e, viewers);
		
		// process children
		NodeList list = e.getChildNodes ();
		int len = list.getLength ();
		
		for (int i = len - 1; i >= 0; i--)
		{
			Node n = list.item (i);
			if (n instanceof Element)
				addLinks ((Element) n, viewers);
		}
	}
	
	/**
	 * adds a simple link to the xpointers Vector
	 */
	protected void addSimpleLink (Element e, Vector viewers)
	{
		String href = e.getAttribute ("xlink:href");
		String file = XPointer.stripXPointer (href);
		
		if (file.equals (""))
			href = url.toString () + href;
		
		// add to list of viewers that need to be updated
		else if (!viewers.contains (file))
			viewers.addElement (file);
		
		if (!XPointer.getXPointer (href).equals (""))
			xpointers.addElement (new Locator (href, e, xlinks.size (), 1));
	}
	
	/**
	 * adds the locators of an extended link to the xpointers Vector
	 */
	protected void addExtendedLink (Element e, Vector viewers)
	{
		int locatorNum = 1;
		
		NodeList locators = e.getChildNodes ();
		int numlocators = locators.getLength ();
		
		for (int k = 0; k < numlocators; k++)
		{
			Node n = locators.item (k);
			if (n instanceof Element)
			{
				Element f = (Element) n;
				if (f.getAttribute ("xlink:form").equals ("locator"))
				{
					String href = f.getAttribute ("xlink:href");
					String file = XPointer.stripXPointer (href);
					
					if (file.equals (""))
						href = url.toString () + href;
					
					// add to list of viewers that need to be updated
					else if (!viewers.contains (file))
						viewers.addElement (file);
					
					if (!XPointer.getXPointer (href).equals (""))
						xpointers.addElement (new Locator 
							(href, f, xlinks.size (), locatorNum++));
				}
			}
		}
	}
	
	/**
	 * opens the documents in a group link
	 */
	protected void addGroupLink (Element e, Vector viewers)
	{
		NodeList locators = e.getChildNodes ();
		int numlocators = locators.getLength ();
		
		for (int k = 0; k < numlocators; k++)
		{
			Node n = locators.item (k);
			if (n instanceof Element)
			{
				Element f = (Element) n;
				if (f.getAttribute ("xlink:form").equals ("document")) try
				{
					String href = f.getAttribute ("xlink:href");
					String file = XPointer.stripXPointer (href);
					Viewer remote = viewerWithURL (new URL (url, file));
					
					// add its links now, process them later
					// (after all links added)
					if (remote == null)
						ViewerFactory.newDefaultViewer 
							(new URL (url, href), false);
					
					if (!viewers.contains (file))
						viewers.addElement (file);
				}
				catch (Exception ex) {}
			}
		}
	}
	
	/**
	 * Processes link tags in open XML files via the xpointers Vector
	 * by adding HTML anchors as specified by XLinks
	 * (the XLinks could potentially come from all open documents)
	 */
	protected void processLinks () throws IOException
	{
		Progress.setString ("processing xlinks " + url.toString ());
		
		for (int i = 0; i < xpointers.size (); i++)
		{
			Locator l = (Locator) xpointers.elementAt (i);
			String xpointer = XPointer.getXPointer (l.href);
			
			// the xpointer points to this doc, add it
			try
			{
				if (url.equals (new URL (url, XPointer.stripXPointer (l.href))))
				{
					NodeList nodes = new XPointer (doc, l.origin, xpointer)
						.getSelection ();
					int len = nodes.getLength ();
					
					for (int j = 0; j < len; j++)
						addAnchor (nodes.item (j), l.linkNum, l.locatorNum);
					
					xpointers.removeElementAt (i);
					i--;
				}
			}
			catch (MalformedURLException murle) {}
		}
	}
	
	/**
	 * Adds a new anchor tag in front Node n
	 * @param n Node the element in front of which the anchor is added
	 * @param link int the link vector index
	 * @param locator int the anchor instance index
	 */
	protected void addAnchor (Node n, int link, int locator)
	{
		Element anchor = doc.createElement ("html:a");
		anchor.setAttribute ("href", "http://xlink/~" + link + "-" + locator);
		anchor.setAttribute ("name", link + "-" + locator);
		
		if (n.hasChildNodes ())
			n.insertBefore (anchor, n.getFirstChild ());
		else
			n.appendChild (anchor);
	}
    
    /**
	 * displays this page
	 */
	public void displayPage ()
	{
		Component window = Function.getRootContainer (this);
		if (window instanceof Window)
			((Window) window).show ();
		
		try
		{
			// make getPage() return a null URL
			// to coerce JEditorPane.setPage() into reloading the page
			getNullPage = true;
			
			super.setPage (url);
			getNullPage = false;
		}
		catch (Exception e)
		{
			Progress.setString ("failed to open " + url);
			e.printStackTrace ();
		}
	}
	
	/**
	 * Processes this Document's stylesheet
	 */
	protected void processStylesheet () throws IOException
	{
		// if we haven't found and modified the stylesheet yet
		if (stylesheet.equals (""))
		{
			stylesheet = Function.userDirToURL ("data/default.xsl");
			
			// override the empty stylesheet if we have local copy
			// of the stylesheet that corresponds to the root element
			// (doc.getDoctype () returns null with the DOM I'm using)
			try
			{
				String local = "data/" + doc.getDocumentElement ().getTagName () 
							 + ".xsl";
				InputStream in = new FileInputStream (local);
				in.close ();
				stylesheet = Function.userDirToURL (local);
			}
			catch (Exception e) {}
			
			// override default with xml-stylesheet pi in this Document
			NodeList list = doc.getChildNodes ();
			for (int i = 0; i < list.getLength (); i++)
			{
				Node n = list.item (i);
				if (n instanceof ProcessingInstruction)
				{
					ProcessingInstruction pi = (ProcessingInstruction) n;
					
					if (pi.getTarget ().equals ("xml-stylesheet"))
					{
						String data = pi.getData ();
						
						// make sure stylesheet is xsl
						if (Function.extractAttrValue (data, "type", "text/xsl")
							.equals ("text/xsl"))
						{
							String remote = Function.extractAttrValue 
									(data, "href", "");
							
							// put together full path (if relative to xml doc)
							if (!remote.startsWith ("http:") &&
								!remote.startsWith ("file:"))
							{
								String path = url.toString ();
								
								remote = path.substring 
									(0, path.lastIndexOf ('/') + 1) + remote;
							}
							
							// make sure the xml-stylesheet derived url exists
							try
							{
								InputStream in = new URL (remote).openStream ();
								in.close ();
								stylesheet = remote;
							}
							catch (IOException ioe) {}
						}
					}
				}
			}
			
			// create & modify copied stylesheet
			createStylesheet (stylesheet, number);
			stylesheet = "data/temp/stylesheet" + number + ".xsl";
		}
		
		// write the processed doc so xt can process it
		Progress.setString ("copying xml file " + url.toString ());
		OutputStream o = new FileOutputStream 
				("data/temp/source" + number + ".xml");
		writeHeader (o);
		dom.writeDocument (doc, o);
		o.close ();
		
		Progress.setString ("converting to html " + url.toString ());
		SAXParser.setBaseURL (url);
		
		// run xt Driver
		String[] args = new String[3];
		args[0] = "data/temp/source" +  number + ".xml";
		args[1] = new String (stylesheet);
		args[2] = "data/temp/result" + number + ".html";
		Driver.main (args);
		
	}
	
	/**
	 * Writes header (processing instructions & DTD) so dom write can do rest
	 */
	protected void writeHeader (OutputStream o) throws IOException
	{
		InputStream i = url.openStream ();
		
		int c = i.read ();
		
		// do until end of file
		while (c != -1)
		{
			// stop once we get a tag that is not a processing instruction <?
			// and not a dtd tag <!
			if (c == '<')
			{
				c = i.read ();
				
				// if we've hit the first regular element tag
				// then we're done w/the header
				if (c != '!' && c != '?')
					break;
				
				o.write ('<');
			}
			
			o.write (c);
			c = i.read ();
		}
		
		i.close ();
		o.flush ();
	}
	
	/**
	 * Adds xsl:apply-templates tags with select="html:a"
	 * for each template without a general xsl:apply-templates child element
	 * which matches our xlinked nodes
	 */
	protected void createStylesheet (String url, int number) throws IOException
	{
		Progress.setString ("processing stylesheet " + url);
		
		Document sheet = dom.openDocument (url);
		if (sheet == null)
			throw new IOException (dom.getException ().toString ());
		
		Element docElem = sheet.getDocumentElement ();
		
		NodeList templates = docElem.getElementsByTagName ("xsl:template");
		int tlen = templates.getLength ();
		
		// check every xsl:template element for xsl:apply-templates elements
		for (int i = 0; i < tlen; i++)
		{
			Element t = (Element) templates.item (i);
			addApplyTemplates (sheet, t);
		}
		
		NodeList fors = docElem.getElementsByTagName ("xsl:for-each");
		int flen = fors.getLength ();
		
		// check every xsl:template element for xsl:apply-templates elements
		for (int i = 0; i < flen; i++)
		{
			Element f = (Element) fors.item (i);
			addApplyTemplates (sheet, f);
		}
		
		// add html:a template rule
		Element img = sheet.createElement ("html:img");
		img.setAttribute ("src", Function.userDirToURL ("data/link.gif"));
		img.setAttribute ("border", "1");
		
		Element a = sheet.createElement ("html:a");
		a.setAttribute ("href", "{@href}");
		a.setAttribute ("name", "{@name}");
		a.appendChild (img);
		
		Element template = sheet.createElement ("xsl:template");
		template.setAttribute ("match", "html:a");
		template.appendChild (a);
		docElem.appendChild (template);
		
		Progress.setString ("copying stylesheet " + url);
		OutputStream o = new FileOutputStream 
			("data/temp/stylesheet" + number + ".xsl");
		dom.writeDocument (sheet, o);
		o.close ();
	}
	
	/**
	 * Adds a xsl:apply-templates tag with select="html:a"
	 * for an Element e without a general xsl:apply-templates child element
	 */
	protected void addApplyTemplates (Document sheet, Element e)
	{
		NodeList applies = e.getElementsByTagName ("xsl:apply-templates");
		int alen = applies.getLength ();
		boolean hasGeneral = false;
		
		for (int i = 0; i < alen; i++)
		{
			Element a = (Element) applies.item (i);
			
			if (a.getAttribute ("select").equals (""))
			{
				hasGeneral = true;
				
				// make sure we didn't find the apply-templates
				// of a nested for-each block
				a = (Element) a.getParentNode ();
				while (a != e)
				{
					if (a.getTagName ().equals ("xsl:for-each"))
					{
						hasGeneral = false;
						break;
					}
					a = (Element) a.getParentNode ();
				}
			}
		}
		
		// if this template has no general xsl:apply-templates element
		// add a xsl:apply-templates select="html:a" element
		if (hasGeneral == false)
			insertApplyTemplates (sheet, e);
	}
	
	/**
	 * Adds a xsl:apply-templates tag with select="html:a"
	 * directly before the first apply-templates, value-of, or text element
	 * not within a for-each, if, or choose element in Element e
	 */
	protected boolean insertApplyTemplates (Document sheet, Element e)
	{
		NodeList list = e.getChildNodes ();
		int len = list.getLength ();
		
		for (int i = 0; i < len; i++)
		{
			Node n = list.item (i);
			
			if (n instanceof Element)
			{
				Element f = (Element) n;
				String tag = f.getTagName ();
				
				// ignore children of these elements
				if (!(tag.equals ("xsl:for-each")
					|| tag.equals ("xsl:if") || tag.equals ("xsl:choose")))
					
					// add apply-templates in front of these elements
					if (tag.equals ("xsl:apply-templates")
						|| tag.equals ("xsl:value-of") || tag.equals ("xsl:text"))
					{
						Element apply = sheet.createElement 
							("xsl:apply-templates");
						apply.setAttribute ("select", "html:a");
						e.insertBefore (apply, f);
						return true;
					}
					// else try the children
					else if (insertApplyTemplates (sheet, f))
						return true;
			}
		}
		
		return false;
	}
	
	/**
	 * makes a menu item for this popup
	*/
	protected void addPopupItem (String name, String command)
	{
		MenuItem i = new MenuItem (name);
		
		i.setActionCommand (command);
	   	i.addActionListener (this);
	   	
	   	popup.add (i);
	}
	
	
	/**
	 * inner class Locator for xpointers array
	 */
	protected class Locator
	{
		public String href;
		public Element origin;
		public int linkNum;
		public int locatorNum;
		
		public Locator (String href, Element origin, int linkNum, int locatorNum)
		{
			this.href = href;
			this.origin = origin;
			this.linkNum = linkNum;
			this.locatorNum = locatorNum;
		}
	}
}