// ViewerFactory.java
// Justin Ludiwg 1999.03.01

package edu.wooster.lu99jp26.link;

import edu.wooster.lu99jp26.util.Function;
import java.awt.Window;
import java.lang.Exception;
import java.net.URL;

public class ViewerFactory
{
	protected static ViewerFactory factory = new ViewerFactory ();
	
	
	/**
	 * gets the current ViewerFactory object
	 * @return ViewerFactory
	 */
	public static ViewerFactory getFactory ()
	{
		return factory;
	}
	
	/**
	 * creates a new Viewer in a ViewerWindow
	 * @return Viewer the new Viewer
	 */
	public static Viewer newDefaultViewer ()
	{
		return getFactory ().newViewer ();
	}
	
	/**
	 * creates a new Viewer in a ViewerWindow
	 * if Viewer with same url already exists, uses it instead
	 * @param url URL the url for the Viewer to display
	 * @return Viewer the new Viewer
	 */
	public static Viewer newDefaultViewer (URL url)
	{
		return getFactory ().newViewer (url.toString (), true);
	}
	
	/**
	 * creates a new Viewer in a ViewerWindow
	 * if Viewer with same url already exists, uses it instead
	 * @param url String the url for the Viewer to display
	 * @return Viewer the new Viewer
	 */
	public static Viewer newDefaultViewer (String url)
	{
		return getFactory ().newViewer (url, true);
	}
	
	/**
	 * creates a new Viewer in a ViewerWindow
	 * if Viewer with same url already exists, uses it instead
	 * @param url URL the url for the Viewer to display
	 * @param autoDisplay boolean if false, must call displayPage ()
	 * @return Viewer the new Viewer
	 */
	public static Viewer newDefaultViewer (URL url, boolean autoDisplay)
	{
		return getFactory ().newViewer (url.toString (), autoDisplay);
	}
	
	/**
	 * creates a new Viewer in a ViewerWindow
	 * if Viewer with same url already exists, uses it instead
	 * @param url String the url for the Viewer to display
	 * @param autoDisplay boolean if false, must call displayPage ()
	 * @return Viewer the new Viewer
	 */
	public static Viewer newDefaultViewer (String url, boolean autoDisplay)
	{
		return getFactory ().newViewer (url, autoDisplay);
	}
	
	
	
	
	/**
	 * default ViewerFactory
	 */
	public ViewerFactory ()
	{
		
	}
	
	/**
	 * creates a new Viewer in a ViewerWindow
	 * @return Viewer the new Viewer
	 */
	public Viewer newViewer ()
	{
		Viewer v = new Viewer ();
		new ViewerWindow (v);
		
		return v;
	}
	
	/**
	 * creates a new Viewer in a ViewerWindow
	 * if Viewer with same url already exists, uses it instead
	 * @param url URL the url for the Viewer to display
	 * @return Viewer the new Viewer
	 */
	public Viewer newViewer (URL url)
	{
		return newViewer (url.toString (), true);
	}
	
	/**
	 * creates a new Viewer in a ViewerWindow
	 * if Viewer with same url already exists, uses it instead
	 * @param url String the url for the Viewer to display
	 * @return Viewer the new Viewer
	 */
	public Viewer newViewer (String url)
	{
		return newViewer (url, true);
	}
	
	/**
	 * creates a new Viewer in a ViewerWindow
	 * if Viewer with same url already exists, uses it instead
	 * @param url URL the url for the Viewer to display
	 * @param autoDisplay boolean if false, must call displayPage ()
	 * @return Viewer the new Viewer
	 */
	public Viewer newViewer (URL url, boolean autoDisplay)
	{
		return newViewer (url.toString (), autoDisplay);
	}
	
	/**
	 * creates a new Viewer in a ViewerWindow
	 * if Viewer with same url already exists, uses it instead
	 * @param url String the url for the Viewer to display
	 * @param autoDisplay boolean if false, must call displayPage ()
	 * @return Viewer the new Viewer
	 */
	public Viewer newViewer (String url, boolean autoDisplay)
	{
		Viewer v = null;
		
		try
		{
			v = Viewer.viewerWithURL (new URL (XPointer.stripXPointer (url)));
			
			if (v == null)
				v = newViewer ();
			
			v.setPage (url, autoDisplay);
		}
		catch (Exception e) {e.printStackTrace ();}
		
		return v;
	}
}